import numpy as np
import trojai.data
from trojai.data.MemoryBank import DimReductionMemBank
import torch
from trojai.estimators.PytorchClassifier import NetWrapper
from sklearn.decomposition import PCA

class LatentDimReduce:
    def __init__(self, network, latent_layer=-2, device=None, dim_reduce_class=PCA, **dim_reducer_kwargs):
        '''

        :param network: Network with which to extract latents from. Can be Pytorch network, or TrojPytorchClassifier.
        :param latent_layer: The latent layer from which to extract the representations.
        :param device: The device network is on. If None, if a GPU exists, it is used, otherwise the CPU is used.
        :param dim_reduce_class: The uninitialized class used for the dimensionality reduction. It must have a fit_transform method.
        :param dim_reducer_kwargs:`Keyword arguments for dimensionality reducer.
        '''
        if type(network).__name__==TrojPytorchClassifier:
            network = network._model
        self.device = device
        if self.device==None:
            self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        self.network_latent_output = NetWrapper(network, latent_layer)
        self.dim_reducer = dim_reduce_class(**dim_reducer_kwargs)
        self.memory_bank = DimReductionMemBank(size=0)

    def populate_memory(self,loader):
        '''
        #TODO allow for select number of batches.
        Populates the memorybank with latent representations as numpy arrays.
        :param loader: Troj dataloader.
        :return: Populated memory bank.
        '''
        #TODO add tracking (instances added to memory vs. instances left)
        for batch_idx, (data, target, index) in enumerate(loader):
            if type(data) is np.ndarray:
                data = torch.from_numpy(data).to(self.device)
            latents = self.network_latent_output(data)
            latents = latents.detach().cpu().numpy()
            self.memory_bank.update(latents, index)


    def fit_reducer(self):
        '''
        Fits the dimensionality reduction algorithm and transforms the latents in the memory bank.
        :return: The embedded latents and the indices of the corresponding sampels in the dataframe.
        '''
        latents, indices = self.memory_bank.get_all()
        latents = np.asarray(latents)
        reduced = self.dim_reducer.fit_transform(latents)
        return reduced, indices

    def fit_transform(self, loader):
        '''
        Populates the memory bank, then fits the dimensionality reducer and transforms the latents accordingly.
        :param loader: Dataloader
        :return: The embedded latents and the indices of the corresponding sampels in the dataframe.
        '''
        self.populate_memory(loader)
        reduced, indices = self.fit_reducer()
        return reduced, indices

    def run(self, loader, df):
        '''
        Given a dataloader and a dataframe, it populates the memory bank, fits and transforms the data, then adds the
        new representations to the dataframe accordingly.
        :param loader: Dataloader.
        :param df: Dataframe to lof the data to.
        :return: The updated dataframe.
        '''
        reduced, indices = self.fit_transform(loader)
        as_dict = {'embedding':reduced}
        df = trojai.data.log_to_dataframe(df, indices, as_dict)
        return df


import pandas as pd
import numpy as np
import plotly.graph_objects as go

import plotly.express as px
from ..plots import *
from ..data import *
from ..metrics import *
from ..metrics.metric_utils import *
from sklearn.metrics import accuracy_score, recall_score, precision_score, f1_score
from collections import defaultdict

#TODO document

def SecurityCurveClassification(df, thresholds = None, classwise = False):
    '''
    Generates the security curve for a given class, or the average curve over all classes.
    :param df: dataframe.
    :param thresholds: Thresholds for calculating security curve. If None, all unique $L^\infty$ values in the dataframe are used.
    :param classwise: Whether or not we are using the function for classwise security curve or not.
    :return: Plotly figure.
    '''
    if thresholds is None:
        thresholds = sorted(df['Linf_perts'].unique().tolist())
    thresholds, accs = GetSCData(df, thresholds)
    if classwise:
        class_name = df['class_name'].unique().tolist()

        fig = go.Scatter(x=thresholds, y=accs,
                        mode='lines',
                        name=class_name[0])
    else:
        plot_data = {'Minimum Perturbation':thresholds, 'Accuracy at Perturbation Amount':accs}
        fig = px.line(plot_data, x="Minimum Perturbation", y="Accuracy at Perturbation Amount", title='Security Curve')
    return fig



def PertLossCorrelationClassification(df, thresholds=None):
    '''
    Generates a curve showing the correlation between the loss and the perturbation amount.
    :param df: Dataframe
    :param thresholds: $L^\infty$ Thresholds for calculating the perturbation-loss correlation curve. If None, all unique $L^\infty$ values in the dataframe are used.
    :return: plotly figure.
    '''
    if thresholds is None:
        thresholds = sorted(df['Linf_perts'].unique().tolist())
    thresholds, accs = GetLossPertCorrelationData(df, thresholds)
    plot_data = {'Minimum Perturbation': thresholds, 'Average Loss for Samples Requiring x Perturbation': accs}
    fig = px.line(plot_data, x="Minimum Perturbation",
                  y="Average Loss for Samples Requiring x Perturbation", title='Perturbation-Loss Curve')
    return fig


def ClasswiseSecurityCurve(df):
    '''
    Generates the classwise security curve.
    :param df: dataframe.
    :return: A plotly figure.
    '''
    lt = plotly.graph_objects.Layout(title='Classwise Security Curve',  xaxis_title="Minimum Perturbation",
        yaxis_title="Accuracy at Perturbation Amount")
    curve = ClasswisePlots(df, SecurityCurveClassification, go_kwargs={'layout':lt}, callback_kwargs={'classwise':True})
    return curve

def ODNormalizedFNGraph(df, ground_truth_df, predictions_df, adv_predictions_df, id_dictionary):
    '''
    A bar chart displaying the false negatives per class, normalized between [0,1] (normal and adversarial).
    :param df: Dataframe
    :param ground_truth_df: Dataframe containing the ground truth objects (can be obtained by computing the mAP score using trojai.metrics.mAP).
    :param predictions_df: Dataframe containing the predicted objects (can be obtained by computing the mAP score using trojai.metrics.mAP).
    :param adv_predictions_df: Dataframe containing the predicted objects under adversarial perturbations
     (can be obtained by computing the mAP score using trojai.metrics.mAP after the adversarial evaluation by setting the adversarial argument in the mAP function to True).
    :param id_dictionary: A dictionary containing for each class the correspondence {class_id:class_name}.
    :return: A plotly figure.
    '''
    false_n = GetFalseNegatives(ground_truth_df, predictions_df)
    label_occurance_counts = ground_truth_df['label_id'].value_counts()
    loc_dict = SwapKeys(label_occurance_counts.to_dict(), id_dictionary)
    false_negs_counts = false_n['label_id'].value_counts()
    fnc_dict = SwapKeys(false_negs_counts.to_dict(), id_dictionary)
    adv_false_n = GetFalseNegatives(ground_truth_df, adv_predictions_df)
    adv_false_negs_counts = adv_false_n['label_id'].value_counts()
    adv_fnc_dict = SwapKeys(adv_false_negs_counts.to_dict(), id_dictionary)
    adv_normed = sort_dict(normalize_dicts(adv_fnc_dict, loc_dict))
    plot_data = []
    plot_data.append(go.Bar(x=list(normed.keys()), y=list(normed.values()), name="Normal"))
    plot_data.append(go.Bar(x=list(adv_normed.keys()), y=list(adv_normed.values()), name="Adversarial"))
    fig = go.Figure(
        data=plot_data,
        layout=dict(title=dict(text="Normalized False Negative Counts By Class"))
    )
    return fig



def ODClassPrecisionGraph(df, predictions_df, adv_predictions_df, id_dictionary):
    '''
    Makes a barchart displaying the classwise precision (normal and adversarial).
    :param df: Dataframe.
    :param predictions_df: Dataframe containing the predicted objects (can be obtained by computing the mAP score using trojai.metrics.mAP).
    :param adv_predictions_df: Dataframe containing the predicted objects under adversarial perturbations
     (can be obtained by computing the mAP score using trojai.metrics.mAP after the adversarial evaluation by setting the adversarial argument in the mAP function to True).
    :param id_dictionary: A dictionary containing for each class the correspondence {class_id:class_name}.
    :return: A plotly figure.
    '''
    class_prec = ClasswisePrecision(predictions_df, id_dictionary)
    adv_class_prec = ClasswisePrecision(adv_predictions_df, id_dictionary)
    plot_data = []
    plot_data.append(go.Bar(x=list(class_prec.keys()), y=list(class_prec.values()), name="Normal Precision"))
    plot_data.append(go.Bar(x=list(adv_class_prec.keys()), y=list(adv_class_prec.values()), name="Adversarial Precision"))
    fig = go.Figure(
        data=plot_data,
        layout=dict(title=dict(text="Classwise Precision"))
    )
    return fig

#TODO move to metrics


def ODClassRecallGraph(df, ground_truth_df, predictions_df, adv_predictions_df, id_dictionary):
    '''
    Makes a barchart for the classwise recall (normal and adversarial).
    :param df: Dataframe.
    :param ground_truth_df: Dataframe containing the ground truth objects (can be obtained by computing the mAP score using trojai.metrics.mAP).
    :param predictions_df:  Dataframe containing the predicted objects (can be obtained by computing the mAP score using trojai.metrics.mAP).
    :param adv_predictions_df: Dataframe containing the predicted objects under adversarial perturbations.
    :param id_dictionary: id_dictionary: A dictionary containing for each class the correspondence {class_id:class_name}.
    :return: a plotly figure.
    '''
    class_rec = ClasswiseRecall(ground_truth_df, predictions_df, id_dictionary)
    adv_class_rec = ClasswiseRecall(ground_truth_df, adv_predictions_df, id_dictionary)
    plot_data = []
    plot_data.append(go.Bar(x=list(class_rec.keys()), y=list(class_rec.values()), name="Normal Recall"))
    plot_data.append(go.Bar(x=list(adv_class_rec.keys()), y=list(adv_class_rec.values()), name="Adversarial Recall"))
    fig = go.Figure(
        data=plot_data,
        layout=dict(title=dict(text="Classwise Recall"))
    )
    return fig


def ODClassF1Graph(df, ground_truth_df, predictions_df, adv_predictions_df, id_dictionary):
    '''
    Makes a barchart for the classwise F1 (normal and adversarial).
    :param df: Dataframe.
    :param ground_truth_df: Dataframe containing the ground truth objects (can be obtained by computing the mAP score using trojai.metrics.mAP).
    :param predictions_df:  Dataframe containing the predicted objects (can be obtained by computing the mAP score using trojai.metrics.mAP).
    :param adv_predictions_df: Dataframe containing the predicted objects under adversarial perturbations.
    :param id_dictionary: id_dictionary: A dictionary containing for each class the correspondence {class_id:class_name}.
    :return: a plotly figure.
    '''
    class_f1 = ClasswiseF1(ground_truth_df, predictions_df, id_dictionary)
    adv_class_f1 = ClasswiseF1(ground_truth_df, adv_predictions_df, id_dictionary)
    plot_data = []
    plot_data.append(go.Bar(x=list(class_f1.keys()), y=list(class_f1.values()), name="Normal F1"))
    plot_data.append(go.Bar(x=list(adv_class_f1.keys()), y=list(adv_class_f1.values()), name="Adversarial F1"))
    fig = go.Figure(
        data=plot_data,
        layout=dict(title=dict(text="Classwise F1"))
    )
    return fig

def ODClassAPGraph(df, class_ap, adv_class_ap, id_dictionary):
    '''
    Makes a barchart for the classwise average precision (normal and adversarial).
    :param df: Dataframe
    :param class_ap: The average precision dictionary (can be obtained by computing the mAP score using trojai.metrics.mAP).
    :param adv_class_ap: The adversarial average precision dictionary (can be obtained by computing the mAP score using trojai.metrics.mAP).
    :param id_dictionary: A dictionary containing for each class the correspondence {class_id:class_name}.
    :return: a plotly figure.
    '''
    class_ap = SwapKeys(class_ap, id_dictionary)
    adv_class_ap = SwapKeys(adv_class_ap, id_dictionary)

    plot_data = []
    plot_data.append(go.Bar(x=list(class_ap.keys()), y=list(class_ap.values()), name="Normal AP"))
    plot_data.append(go.Bar(x=list(adv_class_ap.keys()), y=list(adv_class_ap.values()), name="Adversarial AP"))
    fig = go.Figure(
        data=plot_data,
        layout=dict(title=dict(text="Classwise AP"))
    )
    return fig


def MakeTableFigure(row_names, column_names, vals):
    '''
    Makes a table which can be displayed as a plotly figure goven the appropraie inputs. For example, this function would normally take in data like

    fake_rows = ['row_1', 'row_2']
    fake_columns = [['column_1'], ['column_2'], ['column_3'], ['column_3']]
    fake_values = [[1,2], [3,4], [5,6]]

    :param row_names: list of names of the rows in the table.
    :param column_names: A list containing the names of columns (each must be contained itself in a list)
    :param vals: The values for the table given as a list where each entry in the list is itself a list containing the values
    for each row.
    :return: A plotly figure.
    '''
    header_dict = dict(
        values=column_names,
        line_color='darkslategray',
        fill_color='royalblue',
        align=['left', 'center'],
        font=dict(color='black', size=15),
        height=40)

    row_data = vals.insert(0, row_names)

    cell_data = dict(
        values=vals,
        line_color='darkslategray',
        fill=dict(color=['paleturquoise', 'white']),
        align=['left', 'center'],
        font_size=12,
        height=30)

    plot_data = [go.Table(
        header=header_dict,
        cells=cell_data)]

    fig = go.Figure(plot_data)
    return fig


def ODAdversarialTable(df, mean_ap, adv_mean_ap, class_avp, adv_class_avp, id_dictionary, model_name='model'):
    '''
    Generates a table from data obtained via the adversarial evaluation for object detection.
    :param df: Dataframe
    :param mean_ap: Mean average precision score (at 0.5).
    :param adv_mean_ap: Adversarial mean average precision (at 0.5).
    :param class_avp: The average precision dictionary (can be obtained by computing the mAP score using trojai.metrics.mAP).
    :param adv_class_avp: The adversarial average precision dictionary (can be obtained by computing the mAP score using trojai.metrics.mAP).
    :param id_dictionary:  A dictionary containing for each class the correspondence {class_id:class_name}.
    :param model_name: Name of the model to be displayed in the table.
    :return: A plotly figure.
    '''
    stats_dict = APStats(class_avp, id_dictionary, adv_class_avp)
    stats_dict['mAP@0.5'] = round(mean_ap, 2)
    stats_dict['Adversarial mAP@0.5'] = round(adv_mean_ap,2)
    column_names = list(stats_dict.keys())
    column_names.insert(0, 'Model Name')
    row_name = [model_name]
    values = list(stats_dict.values())
    fig = MakeTableFigure(row_name, column_names, values)
    return fig

def CorruptionAccuracyGraph(df):
    '''
    Makes a barchart for the different corruption accuracy scores.
    :param df: Dataframe.
    :return: A barchart displaying the accuracy on different corruptions.
    '''
    corrupt_acc_dict, diff_dict, clean = PerCorruptionMetric(df, accuracy_score)
    corrupt_acc_dict['none'] = clean
    plot_data = []
    plot_data.append(go.Bar(x=list(corrupt_acc_dict.keys()), y=list(corrupt_acc_dict.values()), name="Accuracy per Corruption"))
    fig = go.Figure(
        data=plot_data,
        layout=dict(title=dict(text="Accuracy per Corruption"))
    )
    return fig

def CorruptionRecallGraph(df):
    '''
    Makes a barchart for the different corruption recalls.
    :param df: Dataframe.
    :return: A barchart displaying the recall on different corruptions.
    '''
    corrupt_acc_dict, diff_dict, clean = PerCorruptionMetric(df, recall_score,  **{'average':'macro'})
    corrupt_acc_dict['none'] = clean
    plot_data = []
    plot_data.append(go.Bar(x=list(corrupt_acc_dict.keys()), y=list(corrupt_acc_dict.values()), name="Recall per Corruption"))
    fig = go.Figure(
        data=plot_data,
        layout=dict(title=dict(text="Recall per Corruption"))
    )
    return fig


def CorruptionPrecisionGraph(df):
    '''
    Makes a barchart for the different corruption precisions.
    :param df: Dataframe.
    :return: A barchart displaying the precision on different corruptions.
    '''
    corrupt_acc_dict, diff_dict, clean = PerCorruptionMetric(df, precision_score,  **{'average':'macro'})
    corrupt_acc_dict['none'] = clean
    plot_data = []
    plot_data.append(go.Bar(x=list(corrupt_acc_dict.keys()), y=list(corrupt_acc_dict.values()), name="Precision per Corruption"))
    fig = go.Figure(
        data=plot_data,
        layout=dict(title=dict(text="Precision per Corruption"))
    )
    return fig



def CorruptionF1Graph(df):
    '''
    Makes a barchart for the different corruption F1 scores.
    :param df: Dataframe.
    :return: A barchart displaying the F1 scors on different corruptions.
    '''
    corrupt_acc_dict, diff_dict, clean = PerCorruptionMetric(df, f1_score,  **{'average':'macro'})
    corrupt_acc_dict['none'] = clean
    plot_data = []
    plot_data.append(go.Bar(x=list(corrupt_acc_dict.keys()), y=list(corrupt_acc_dict.values()), name="F1 per Corruption"))
    fig = go.Figure(
        data=plot_data,
        layout=dict(title=dict(text="F1 per Corruption"))
    )
    return fig

def Loss_ecdf(df):
    '''
    Generates the empirical cumulative distribution of the loss for each corruption.
    :param df: Dataframe
    :return: The loss eCDF.
    '''
    fig = px.ecdf(df, x="Loss", color="corruption_type", title='Empirical Cumulative Distribution of the Loss for Each Corruption')
    return fig


def MakeCorruptionsTable(df):
    '''
    Makes a table containing information about performance on different corruptions.
    :param df: Dataframe.
    :return: Table of information.
    '''
    corruptions = df['corruption_type'].unique()
    column_names = [['Corruption'], ['Loss'], ['Accuracy'], ['Precision'], ['Recall'], ['F1']]
    stats_dict = defaultdict(list)
    for corrupt in corruptions:
        corrupt_df = df[df['corruption_type'] == corrupt]
        loss_str = '{}+/-{}'
        loss_mean, loss_std = ColumnMeanStd(corrupt_df, 'Loss')
        stats_dict['Loss'].append(loss_str.format(loss_mean, loss_std))
        stats_dict['Accuracy'].append(round(ComputeMetric(corrupt_df, accuracy_score), 3))
        stats_dict['Precision'].append(round(ComputeMetric(corrupt_df, precision_score, **{'average': 'macro'}), 3))
        stats_dict['Recall'].append(round(ComputeMetric(corrupt_df, recall_score, **{'average': 'macro'}), 3))
        stats_dict['F1'].append(round(ComputeMetric(corrupt_df, f1_score, **{'average': 'macro'}), 3))
    vals = list(stats_dict.values())
    fig = MakeTableFigure(corruptions, column_names, vals)
    return fig



# import torch
# import torch.nn as nn
import tensorflow as tf

# class PytorchObjectDetector(nn.Module):
#     def __init__(self, model):
#         """

#         :param model: Pytorch Object Detection model
#         """
#         super(PytorchObjectDetector, self).__init__()
#         self.model = model

#     def compute_loss(self, samples, labels, grad=False, reduce=True):
#         """
#         :param: samples: Samples to make predictions on as list of tensors.
#         :param: labels: labels as list of dictionaries
#         :param: grad: If true, the gradient is computed.
#         :param: reduce: If true, the losses are reduced via summation and a single scalar is returned, otherwise a dictionary is returned.
#         """
#         self.model.train()
#         if grad == False:
#             with torch.no_grad():
#                 losses = self.model(samples, labels)
#         else:
#             losses = self.model(samples, labels)
#         if reduce:
#             losses = sum(loss for loss in losses.values())
#         return losses

#     def predict(self, sample):
#         self.model.eval()
#         preds = self.model(sample)
#         return preds



from art import estimators



"""
This class is our base class that child classes will inherit metadata collection (and other functions in the future) as well as inherits from
ART's base estimator class for instantiation and attack. 
We'll have a class for major frameworks that will inherit this class

"""


class TrojClassifier(estimators.BaseEstimator):
    """
    Collects and returns the model's metadata
    :return dict: returns a dictionary of metadata related to the classifier passed to the ART estimator class
    """

    def get_classifier_meta(self):
        return {
            "model": str(self._model),
            "model_name": str(self._model.__class__.__name__),
            "input_shape": str(self._input_shape),
            "num_classes": str(self.nb_classes),
            "loss_func": str(self._loss),
            "channels_first": str(self._channels_first),
        }








from trojai import client
import numpy as np
import json
from trojai.logger import logger
from trojai.client.client_utils import save_dataframe_to_disk


class TrojSession:
    def __init__(self):
        super().__init__()
        self.client = None
        self.logger = logger.Logger()
        self.graphs = None

    def run_name(self, name):
        self.log({"run_name": name})

    def log_graph(self, html):
        self.log({"graphs": html})

    def log(self, in_dict):
        self.logger.log(in_dict)

    def create_project(self, project_name: str):
        return self.client.create_project(project_name)

    def create_dataset(self, project_name: str, dataset_name: str):
        return self.client.create_dataset(project_name, dataset_name)

    def reorient_df(self, dataframe):
        dataframe = json.loads(dataframe.to_json(orient="index"))
        return dataframe

    def _check_df_storage(self):
        # Check user/teams datapoint limit and actual storage
        datapoints_left = self.client.check_storage()
        logger_dict = self.logger.get_logger()

        if "dataframe" in logger_dict:
            return len(logger_dict["dataframe"]) <= datapoints_left
        # Raise exception here instead if "dataframe" isn't in dict?
        return True

    """
    Drops rows with a null value in a column from the dataframe then reorients the dataframe to be returned
    :param dataframe: the dataframe
    :param drop_na: nulls should be dropped bool
    """

    def drop_empty_rows(self, dataframe, drop_na):
        if drop_na == True:
            dataframe = dataframe.dropna()

        json_processed_out = self.reorient_df(dataframe)

        return json_processed_out

    """
    Uploads the logs collected by the logger

    :param project_name: string name of project that exusts on the platform
    :param dataset_name: a dataset that exists under the project  
    """

    def upload_logs(self, project_name: str, dataset_name: str, drop_na=True):
        logger_dict = self.logger.get_logger()
        processed_df = logger_dict

        if "metadata" in processed_df:
            processed_df["metadata"] = processed_df["metadata"][0]
        if "dataframe" in processed_df:
            processed_df["dataframe"] = self.drop_empty_rows(
                logger_dict["dataframe"][0], drop_na
            )
        jsonified_df = processed_df

        # Save dataframe to disk before uploading to s3
        # save_dataframe_to_disk(dataset_name, jsonified_df)

        # Check if dataframe will be able to be uploaded based on current storage usage
        if not self._check_df_storage():
            raise Exception(
                "This upload will exceed datapoint limit. Delete existing datasets or upgrade your subscription to gain access to more datapoint storage."
            )

        return self.client.upload_df_results(project_name, dataset_name, jsonified_df)

    def add_latent(self, np_latent, dataframe):
        dataframe["latent"] = np_latent
        return dataframe



    """
    Takes in the troj dataframe we've built over the attack loop as well as optionally the classifier, evaluator and dataloader
    for metadata collection. This function must be run before the upload_logs function for proper formatting.

    :param dataframe: the dataframe used to track the experiments and data
    :param classifier: the TrojClassifier object
    :param evaluator: The RobustnessEvaluator object
    :param dataloader: the dataloader object  
    :param tags: array of strings to tage the run 
    :return: dictionary of metadata
    """

    def log_metadata(
        self, dataframe, classifier=None, evaluator=None, dataloader=None, tags=[]
    ):
        # also need to pull the log_dict from logger
        classifier_meta = {}
        evaluator_meta = {}
        dataloader_meta = {}

        if classifier is not None:
            classifier_meta = classifier.get_classifier_meta()
        if evaluator is not None:
            evaluator_meta = evaluator.atk_meta
        if dataloader is not None:
            dataloader_meta = dataloader.dataset_meta
        if "prediction" in dataframe:
            dataframe["prediction"].replace("", np.nan, inplace=True)
        dataframe.dropna(inplace=True)
        out_dict = {
            "metadata": {
                "classifier_metadata": classifier_meta,
                "evaluator_metadata": evaluator_meta,
                "dataloader_metadata": dataloader_meta,
                "tags": str(tags),
            },
            "dataframe": dataframe,
        }
        self.log(out_dict)

        return out_dict



from art import estimators
import tensorflow as tf
import numpy as np
from trojai.estimators.ClassifierBase import TrojClassifier

"""
Inherits from the troj base class and the ART tensorflow v2 classifier for tf2 support
Also implements our ComputeLoss function which will be called by each classifier, should implement as abstract method in base class
"""


class TrojKerasClassifier(estimators.classification.KerasClassifier, TrojClassifier):
    import tensorflow as tf

    tf.compat.v1.disable_eager_execution()

    """
    Compute loss computes the loss over a batch of target samples

    :param x: inputs
    :param y: labels, either true labels or original unperturbed model labels. y might need to be expanded along
    the first dimension because of art bug.
    :param return_preds: bool to return model predictions with loss
    :param reduction: changes the reduction of the loss function to none 
    :return: Loss values and optionally predictions
    """

    def ComputeLoss(
        self, x, y, return_preds=True, reduction=tf.keras.losses.Reduction.NONE
    ):
        # import tensorflow.compat.v1 as tf
        import tensorflow as tf

        old_reduction = self.model.loss_functions[0]._get_reduction()
        self.model.loss_functions[0].reduction = reduction
        preds = self.predict(x)
        # hard_labels = np.argmax(preds, axis=1)
        # hard_labels_dtype = 'float32'
        # y = y.astype(hard_labels_dtype)
        loss_val = self.model.loss_functions[0](y, preds)
        self._loss.reduction = old_reduction
        # tf.compat.v1.enable_eager_execution()
        if return_preds:

            return loss_val.eval(session=tf.compat.v1.Session()), preds
        else:
            return loss_val.numpy()


from art import estimators
import numpy as np
from trojai.estimators.ClassifierBase import TrojClassifier
try:
    import tensorflow.compat.v1 as tf
except:
    import tensorflow as tf
"""
Inherits from the troj base class and the ART tensorflow v1 classifier for tf1 support
Also implements our ComputeLoss function which will be called by each classifier, should implement as abstract method in base class
"""


class TrojTF1Classifier(estimators.classification.TensorFlowClassifier, TrojClassifier):
    from art.estimators.classification import TensorFlowClassifier

    """
    Compute loss computes the loss over a batch of target samples
    :param x: inputs
    :param y: labels, either true labels or original unperturbed model labels. y might need to be expanded along
    the first dimension because of art bug.
    :param return_preds: bool to return model predictions with loss
    :return: Loss values and optionally predictions
    """

    def ComputeLoss(self, x, y, return_preds=True):
        try:
            import tensorflow.compat.v1 as tf
        except:
            import tensorflow as tf
        from numpy import array

        # this has got to be wrong
        shape = (y.size, self.nb_classes)
        one_hot = np.zeros(shape)
        losses = np.array([])
        for sample in x:
            image = np.expand_dims(sample, axis=0)
            rows = np.arange(y.size)
            # print(rows)
            one_hot[rows, y] = 1.0
            # print(one_hot)
            label = np.expand_dims(one_hot[0], axis=0)
            loss_val = self.compute_loss(x, one_hot)
            losses = np.append(losses, loss_val)

        if return_preds:
            preds = self.predict(x)
            return losses, preds
        else:
            return loss_val.numpy()


import plotly
import pandas as pd
import numpy as np
import plotly.express as px
import plotly.graph_objects as go

def GraphMakerBase(graph_func, df, *args, to_html = True, show=False,  **kwargs):
    '''
    Base class for making graphs from dataframes.
    :param graph_func: A function that returns a plotly figure.
    :param df: A dataframe (or dictionary) passed as the first argument to the graph_func
    :param *args: Arguments to be passed to the graph function.
    :param to_html: Whether or not to return the HTML string.
    :param show: Whether or not to display the graph.
    :param **kwargs: Keyword arguments for the graphing function.
    :returns: Returns the html for teh graph if it exists, and None otherwise.
    '''
    fig = graph_func(df, *args, **kwargs)
    html_str = None
    if to_html:
        html_str = plotly.io.to_html(fig, include_plotlyjs=False, full_html=False)
    if show:
        fig.show()
    return html_str

def ClasswisePlots(df, callback, go_kwargs={}, callback_kwargs={}):
    '''
    Used to make plots with different traces for each class if such functionality is not readily available via plotly.express.
    :param df: Dataframe.
    :param callback: A single callback function which returns a trace object.
    :param go_kwargs: A dictionary of keyword arguments to be passed to the graph object figure.
    :param callback_kwargs: A dictionary of keyword arguments to be passed to the callback function.
    '''
    traces = []
    classes = df['class_name'].unique().tolist()
    for class_name in classes:
        sub_df = df[df['class_name']==class_name]
        graph_object = callback(sub_df, **callback_kwargs)
        traces.append(graph_object)
    fig = go.Figure(data=traces, **go_kwargs)
    return fig


import pandas as pd
import numpy as np
import plotly.graph_objects as go
from ..metrics import *
from ..plots import *
from ..plots.express import *
from ..client.Session import TrojSession
import plotly.express as px
#import htmlmin

#TODO more plots
adver_plot_funcs = [SecurityCurveClassification, ClasswiseSecurityCurve, PertLossCorrelationClassification]
adver_plot_funcs_kwargs = [{}, {}, {}]
#Not clever, but works
def EZAdversarialClassification(df, session, plot_func_list = adver_plot_funcs,
                                plot_func_kwargs_list = adver_plot_funcs_kwargs):
    '''
    Makes basic plots for adversarial attacks on classifiers.
    :param df: Dataframe
    :param session: Troj session object.
    '''
    for idx in range(len(plot_func_list)):
        plot_func = plot_func_list[idx]
        plot_func_kwargs = plot_func_kwargs_list[idx]
        html_str = GraphMakerBase(plot_func, df, **plot_func_kwargs)
        #html_str = htmlmin.minify(html_str, remove_comments=True, remove_empty_space=True)
        session.log_graph(html_str)



def EZAdversarialObjectDetection(df, session, id_dictionary, model_name='model'):
    '''
    Makes basic plots for adversarial attacks on object detectors.
    :param df: Dataframe
    :param session: Troj session object.
    :param id_dictionary: A dictionary containing for each class the correspondence {class_id:class_name}.
    :param model_name: Name of model to be put in the displayed table.
    '''
    map_05, gt_df, preds_df, class_ap = mAP(df)
    adv_map_05, adv_gt_df, adv_preds_df, class_ap_adv = mAP(df, adversarial=True)
    od_adver_plot_funcs = [ODAdversarialTable, ODClassPrecisionGraph,ODClassRecallGraph,ODClassF1Graph, ODClassAPGraph]
    od_adver_plot_funcs_kwargs = [{'mean_ap':map_05, 'adv_mean_ap':adv_map_05, 'class_avp':class_ap, 'adv_class_avp':class_ap_adv, 'id_dictionary':id_dictionary, 'model_name':model_name},
                                  {'predictions_df':preds_df, 'adv_predictions_df':adv_preds_df, 'id_dictionary':id_dictionary},
                                  {'ground_truth_df':gt_df ,'predictions_df':preds_df, 'adv_predictions_df':adv_preds_df,
                                   'id_dictionary':id_dictionary}, {'ground_truth_df':gt_df ,'predictions_df':preds_df, 'adv_predictions_df':adv_preds_df,
                                   'id_dictionary':id_dictionary}, {'class_ap':class_ap, 'adv_class_ap':class_ap_adv, 'id_dictionary':id_dictionary}]


    for idx in range(len(od_adver_plot_funcs )):
        plot_func = od_adver_plot_funcs [idx]
        plot_func_kwargs = od_adver_plot_funcs_kwargs [idx]
        html_str = GraphMakerBase(plot_func, df, **plot_func_kwargs)
        #html_str = htmlmin.minify(html_str, remove_comments=True, remove_empty_space=True)
        session.log_graph(html_str)

def EZClassificationCorruptions(df, session):
    '''
    Makes basic plots for evluating the performance of classification models against different corruptions.
    :param df: Dataframe.
    :param session: Troj session.
    '''
    corrupt_plot_funcs = [MakeCorruptionsTable, Loss_ecdf, CorruptionAccuracyGraph, CorruptionRecallGraph, CorruptionPrecisionGraph,
                          CorruptionF1Graph]
    corrupt_plot_func_kwargs = [{}, {}, {}, {}, {}, {}]
    for idx in range(len(corrupt_plot_funcs)):
        plot_func = corrupt_plot_funcs[idx]
        plot_func_kwargs = corrupt_plot_func_kwargs[idx]
        html_str = GraphMakerBase(plot_func, df, **plot_func_kwargs)
        #html_str = htmlmin.minify(html_str, remove_comments=True, remove_empty_space=True)
        session.log_graph(html_str)



from art import estimators
import torch
import torch.nn as nn
from trojai.estimators.ClassifierBase import TrojClassifier
import torch.nn as nn

"""
Inherits from the troj base class and the ART Pytorch classifier for Pytorch support
Also implements our ComputeLoss function which will be called by each classifier, should implement as abstract method in base class
"""


class TrojPytorchClassifier(
    estimators.classification.PyTorchClassifier, TrojClassifier
):
    """
    Compute loss computes the loss over a batch of target samples

    :param x: inputs
    :param y: labels, either true labels or original unperturbed model labels. y might need to be expanded along
    the first dimension because of art bug.
    :param return_preds: bool to return model predictions with loss
    :return: Loss values and optionally predictions
    """

    def ComputeLoss(self, x, y, return_preds=True, reduction="none"):

        old_reduction = self._loss.reduction
        self._loss.reduction = reduction
        preds = torch.tensor(self.predict(x))
        y = torch.tensor(y)
        loss_val = self._loss(preds, y)
        self._loss.reduction = old_reduction
        if return_preds:
            return loss_val.numpy(), preds.numpy()
        else:
            return loss_val.numpy()


def flatten(t):
    '''

    :param t: Input array of shape [N, d_1, ..., d_m ]
    :return: flattened array of shape [N, d_1*d_2*...d_m]
    '''
    return t.reshape(t.shape[0], -1)

class NetWrapper(nn.Module):
    # Modified from https://github.com/lucidrains/byol-pytorch
    #for extracting latent, layer can either be string or index.

    def __init__(self, net, layer = -2):
        '''
        A wrapper for extracting latent representations from a Pytorch model.
        :param net: Network to pass to the wrapper
        :param layer: Either layer index or layer name to get the latent representations from.b
        '''
        super().__init__()
        if type(net).__name__==TrojPytorchClassifier:
            self.net = net._model
        else:
            self.net = net
        self.layer = layer

        self.hidden = None
        self.hook_registered = False

    def _find_layer(self):
        '''
        Finds the hidden layer given by the layer input argument.
        :return:
        '''
        if type(self.layer) == str:
            modules = dict([*self.net.named_modules()])
            return modules.get(self.layer, None)
        elif type(self.layer) == int:
            children = [*self.net.children()]
            return children[self.layer]
        return None

    def _hook(self, _, __, output):
        '''

        :param _:
        :param __:
        :param output:
        :return:
        '''
        self.hidden = flatten(output)

    def _register_hook(self):
        '''
        Registers the hook given by _hook in the network at the layer given in the input argument.
        :return:
        '''
        layer = self._find_layer()
        assert layer is not None, f'hidden layer ({self.layer}) not found'
        handle = layer.register_forward_hook(self._hook)
        self.hook_registered = True

    def get_representation(self, x):
        '''
        Given an input, return the latent representation of x
        :param x: batched input to network.
        :return: Batch of output representations.
        '''
        if self.layer == -1:
            return self.net(x)

        if not self.hook_registered:
            self._register_hook()

        _ = self.net(x)
        hidden = self.hidden
        self.hidden = None
        assert hidden is not None, f'hidden layer {self.layer} never emitted an output'
        return hidden

    def forward(self, x):
        representation = self.get_representation(x)

        return representation


import numpy as np
from ..metrics.mAP_utils import MakePredAnnotBoxDataframes, box_ious, ClassAP
from ..metrics.metric_utils import *


def mAP(dataframe, iou_thresh=0.5, adversarial=False):
    '''
    Computes the mean average precision of an object detection algorithm from a dataframe, along with other related objects.
    :param dataframe: Dataframe.
    :param iou_thresh: Threshold to use for the mean average precision.
    :param adversarial: If True, the mean average precision will be computed on the adversarial predictions.
    :return: The mean average precision, and two special dataframes. One referred to as the ground_truth_df whose rows contain
    individual ground truth objects, along with their bounding box and what image they belong to. The other dataframe, called
    predictions_df does a similar thing for the predictions made by the model. We also return a dictionary which contains the average precision per class.
    '''
    ground_truth_df, predictions_df = MakePredAnnotBoxDataframes(dataframe, adversarial)
    ground_truth_df, predictions_df = box_ious(ground_truth_df, predictions_df)
    ap_dict = ClassAP(ground_truth_df, predictions_df)
    return np.mean(list(ap_dict.values())), ground_truth_df, predictions_df, ap_dict


def true_min_pert(labels, preds, original, perturbation, losses, norm=np.inf):
    """
    Finds the smallest perturbation which flips the label. If no perturbation is found, sets the minimum to 0.5 (since
    the image will become entirely grey at such a perturbation).

    :param original_ims: Original inputs
    :param outs_array: Adversarial examples
    :param losses: Adversarial losses
    :param preds: Predictions on Adversarial Examples
    :param norm: the L^p norm to use.
    :return: minimum perturbation, the loss for the perturbation, and the prediction
    """

    tile_list = [original for i in range(perturbation.shape[1])]
    tiled = np.stack(tile_list)
    # swap the first and second axis
    tiled = np.swapaxes(tiled, 0, 1)
    if tiled.shape[1] != 1:
        assert tiled[0, 0, 0, 0, 0] == tiled[0, 1, 0, 0, 0]

    # reshape outputs and original images to collection of vectors
    flattened_shape = (
        perturbation.shape[0],
        perturbation.shape[1],
        perturbation.shape[2] * perturbation.shape[3] * perturbation.shape[4],
    )
    flattened_outs = np.reshape(perturbation, flattened_shape)
    flattened_original = np.reshape(tiled, flattened_shape)

    # subtract the original from the perturbed to get the perturbation vector
    perturbations = flattened_outs - flattened_original
    perturbation_norms = np.linalg.norm(perturbations, norm, axis=2)

    tiled_labels = [labels for i in range(preds.shape[1])]
    tiled_labels = np.stack(tiled_labels)
    # swap the first and second axis
    tiled_labels = np.swapaxes(tiled_labels, 0, 1)
    # If the label isn't flipped, set min pert to 0.5
    perturbation_norms[tiled_labels == np.argmax(preds, axis=2)] = 0.5
    min_per_sample_idx = np.argmin(perturbation_norms, axis=1)

    min_pert_losses = []
    min_pert_preds = []
    min_pert_outs = []
    for idx in range(len(min_per_sample_idx)):
        min_pert_outs.append(perturbation[idx, min_per_sample_idx[idx]])
        min_pert_losses.append(losses[idx, min_per_sample_idx[idx]])
        min_pert_preds.append(preds[idx, min_per_sample_idx[idx]])

    min_pert_outs = np.asarray(min_pert_outs)
    min_pert_preds = np.asarray(min_pert_preds)
    min_pert_losses = np.asarray(min_pert_losses)
    return min_pert_outs, min_pert_preds, min_pert_losses


def ComputeAccuracy(df):
    cor_df = df[df['label'] == df['prediction']]
    accuracy = len(cor_df) / len(df)
    return accuracy

def GetSCData(df, thresholds, mode='classification'):
    '''
    Computes data for generating the security curve.
    :param df: Dataframe
    :param thresholds: Thresholds for computing the curve.
    :param mode:
    :return: The thresholds and accuracy at those thresholds as a list.
    '''
    metrics = []
    cor_df = df[df['label'] == df['prediction']]
    metrics.append(len(cor_df) / len(df))
    if mode == 'classification':
        for thresh in thresholds:
            sub_df = cor_df[cor_df['Linf_perts'] <= thresh]
            metrics.append((len(cor_df) - len(sub_df)) / len(df))
    thresholds.insert(0, 0)
    return thresholds, metrics

def GetLossPertCorrelationData(df, thresholds, mode='classification'):
    '''
    Computes data for generating the Perturbation-loss curve.
    :param df: Dataframe
    :param thresholds: Thresholds for computing the curve.
    :param mode:
    :return: The thresholds and loss at those thresholds as a list.
    '''
    metrics = []
    if mode == 'classification':
        for thresh in thresholds:
            sub_df = df[df['Linf_perts'] <= thresh]
            metrics.append(np.mean(sub_df['Loss'].tolist()))
    return thresholds, metrics



def GetFalseNegatives(ground_truth_df, predictions_df):
    '''
    Finds the false negatives from the results of an object detector.
    :param ground_truth_df: The ground truth dataframe (see trojai.metrics.mAP)
    :param predictions_df: The predictions dataframe (see trojai.metrics.mAP)
    :return: A dataframe containing the false negatives.
    '''
    # gets false negatives
    true_positives = predictions_df[predictions_df['gt_instance_idx'] != -1]
    tp_ids = true_positives['gt_instance_idx'].tolist()
    gt_index_list = ground_truth_df.index.tolist()
    false_negs = diff(gt_index_list, tp_ids)
    fn_df = ground_truth_df.loc[false_negs]
    return fn_df

def ClasswisePrecision(predictions_df, id_dictionary):
    '''
    Computes the classwise precision of the results of an object detector.
    :param predictions_df: The predictions dataframe (see trojai.metrics.mAP)
    :param id_dictionary: A dictionary containing for each class the correspondence {class_id:class_name}.
    :return: a dictionary containing the precision of each class.
    '''
    true_positives = predictions_df[predictions_df['gt_instance_idx']!=-1]
    total_preds_counts = SwapKeys(predictions_df['pred_label_id'].value_counts().to_dict(), id_dictionary)
    tp_counts = SwapKeys(true_positives['pred_label_id'].value_counts().to_dict(), id_dictionary)
    class_prec = sort_dict(normalize_dicts(tp_counts, total_preds_counts))
    return class_prec

def ClasswiseRecall(ground_truth_df, predictions_df, id_dictionary):
    '''
    Computes the recall of the results of an object detector.
    :param ground_truth_df: The ground truth dataframe (see trojai.metrics.mAP).
    :param predictions_df: The predictions dataframe (see trojai.metrics.mAP).
    :param id_dictionary: a dictionary containing the precision of each class.
    :return: Recall of each class.
    '''
    true_positives = predictions_df[predictions_df['gt_instance_idx']!=-1]
    tp_counts = SwapKeys(true_positives['pred_label_id'].value_counts().to_dict(), id_dictionary)
    false_n = GetFalseNegatives(ground_truth_df, predictions_df)
    false_negs_counts = false_n['label_id'].value_counts()
    fnc_dict = SwapKeys(false_negs_counts.to_dict(), id_dictionary)
    denom_dict = add_dicts(tp_counts, fnc_dict)
    recall_dict = normalize_dicts(tp_counts, denom_dict)
    return recall_dict

def ClasswiseF1(ground_truth_df, predictions_df, id_dictionary):
    '''
    Computes F1 scores for object detectors.
    :param ground_truth_df: The ground truth dataframe (see trojai.metrics.mAP).
    :param predictions_df: The predictions dataframe (see trojai.metrics.mAP).
    :param id_dictionary: a dictionary containing the precision of each class.
    :return: F1 score for each class.
    '''
    precisions = ClasswisePrecision(predictions_df, id_dictionary)
    recalls = ClasswiseRecall(ground_truth_df, predictions_df, id_dictionary)
    denom = add_dicts(precisions, recalls)
    nume = multiply_dicts(precisions, recalls)
    normed = normalize_dicts(nume, denom)
    f1 = scalar_mult_dict(2, normed)
    return f1


def APStats(ap_dict, id_dictionary, adv_ap_dict=None):
    '''
    Computes statistics related to the classwise average precision.
    :param ap_dict: Dictionary containing the avereage precisions per class.
    :param id_dictionary: a dictionary containing the precision of each class.
    :param adv_ap_dict: Dictionary containing the adversarial avereage precisions per class.
    :return: A dictionary.
    '''
    stats_dict = {}
    ap_dict = SwapKeys(ap_dict, id_dictionary)
    max_ap_class, max_ap = argmax_dict(ap_dict)
    min_ap_class, min_ap = argmin_dict(ap_dict)
    stats_dict['Highest AP Class'] = max_ap_class
    stats_dict['Lowest AP Class'] = min_ap_class
    stats_dict['Highest AP'] = round(max_ap, 4)
    stats_dict['Lowest AP'] = round(min_ap, 4)
    if adv_ap_dict != None:
        adv_ap_dict = SwapKeys(adv_ap_dict, id_dictionary)
        difference = sub_dicts(ap_dict, adv_ap_dict)
        max_ap_class_adv, max_ap_adv = argmax_dict(adv_ap_dict)
        min_ap_class_adv, min_ap_adv = argmin_dict(adv_ap_dict)
        largest_diff_class, big_diff = argmax_dict(difference)
        smallest_diff_class, small_diff = argmin_dict(difference)
        stats_dict['Highest Adversarial AP Class'] = max_ap_class_adv
        stats_dict['Lowest Adversarial AP Class'] = min_ap_class_adv
        stats_dict['Highest AP'] = round(max_ap_adv, 4)
        stats_dict['Lowest AP'] = round(min_ap_adv, 4)
        stats_dict['Highest Difference Class'] = largest_diff_class
        stats_dict['Lowest Difference Class'] = smallest_diff_class
        stats_dict['Highest Difference'] = round(big_diff, 4)
        stats_dict['Lowest Difference'] = round(small_diff, 4)
    return stats_dict

def ComputeMetric(df, callback, method='predictions', **cb_kwargs):
    '''
    Given a dataframe and a function which takes in one of either (labels, predictions) or (labels, probabilities),
    compute the value of the function on the corresponding dataframe columns.

    :param df: Dataframe
    :param callback: The function to compute.
    :param method: Either 'predictions' or 'probabilities'. If predictions, the hard predictions are passed to the
    callback function. If 'probabilites', it is assumed that the 'pred_probs' column in the dataframe contains logits,
    so we compute the softmax for each row in the 'pred_probs' column, which are the values that are then passed to the callback function.
    :param cb_kwargs: Any extra keyword arguments for the callback.
    :return:
    '''
    #if callback expects class probabilites, set method=probabilites
    labels = df['label'].tolist()
    if method=='predictions':
        preds = df['Preds'].tolist()
    elif method=='probabilities':
        #assumes that pred_probs is logits
        preds = scipy.special.softmax(np.asarray(df['pred_probs'].tolist()), 1)
    outs = callback(labels, preds, **cb_kwargs)
    return outs

def ComputeCorruptionMetricDifference(df, callback, corruption_type=None, severity_level=None, method='predictions', **cb_kwargs):
    '''
    Computes the difference in value of the output of a callback function between corrupted and non-corrupted samples.

    :param df: Dataframe.
    :param callback: The function to be computed.
    :param corruption_type: Corruption type. If None, all corruptions are used.
    :param severity_level: Severity level. If None, value is computed over all severity levels.
    :param method: Either 'predictions' or 'probabilities'. If predictions, the hard predictions are passed to the
    callback function. If 'probabilites', it is assumed that the 'pred_probs' column in the dataframe contains logits,
    so we compute the softmax for each row in the 'pred_probs' column, which are the values that are then passed to the callback function.
    :param cb_kwargs: Keyword arguments to the callback function.
    :return: The score on the clean images, the score on the corrupted images, and the difference between the two.
    '''
    clean_df = df[df['corruption_type']=='none']
    if corruption_type==None:
        corrupt_df = df[df['corruption_type']!='none']
    else:
        corrupt_df = df[df['corruption_type']==corruption_type]
    if severity_level != None:
        corrupt_df = corrupt_df[corrupt_df['severity']==severity_level]
    clean_score = ComputeMetric(clean_df, callback, method=method, **cb_kwargs)
    corrupt_score = ComputeMetric(corrupt_df, callback, method=method, **cb_kwargs)
    diff = clean_score - corrupt_score
    return clean_score, corrupt_score, diff

def PerCorruptionMetric(df, callback, method='predictions', **cb_kwargs):
    '''
    Gets difference in corruption callback per corruption.
    :param df: Dataframe.
    :param callback: The function to be computed.
    :param method: Either 'predictions' or 'probabilities'. If predictions, the hard predictions are passed to the
    callback function. If 'probabilites', it is assumed that the 'pred_probs' column in the dataframe contains logits,
    so we compute the softmax for each row in the 'pred_probs' column, which are the values that are then passed to the callback function.
    :param cb_kwargs: Keyword arguments to the callback function.
    :return: Accuracy on corruptions dictionary, difference dictionary, clean accuracy.
    '''
    #Gets difference in corruption callback per corruption
    corruptions = df[df['corruption_type']!='none']['corruption_type'].unique()
    corr_act_dict = {}
    diff_dict = {}
    for corrupt in corruptions:
        clean, corr, differ = ComputeCorruptionMetricDifference(df, callback, corrupt, method=method, **cb_kwargs)
        corr_act_dict[corrupt] = corr
        diff_dict[corrupt] = differ
    return corr_act_dict, diff_dict, clean


import numpy as np
import trojai.data


class RobustnessEvaluatorBase:
    def __init__(self, model, callbacks, callback_kwargs):
        '''
        #TODO make all current evaluators use this base class.
        A base class for robustness evaluators.

        :param model: A placeholder used for the determining the structure of callbacks. If model != None, it is then the
        first argument passed to designated callback functions. Otherwise, the first argument to the callback is expected to be
        the input data.
        :param callbacks: A list of callback functions. If model != None, all functions must accept as the first 4 input arguments model,
        input data to the model, the target, indices and then any callback keyword arguments defined in the callback_kwargs argument.
        If model is None, then the callbacks all take in the data, the target, and indices as the first three arguments.
        Each callback function must return two values: a list of indices, and a dictionary of values to log to the
        dataframe at those indices.
        :param callback_kwargs: A list of keyword arguments for each callback function in callbacks.
        '''
        self.model = model
        self.callbacks = callbacks
        self.callback_kwargs = callback_kwargs

    def run(self, df, loader, num_batches=None, verbose=True):
        #TODO add verbose flag
        '''
        runs the evaluations defined in the callback functions.

        :param df: a dataframe to be logged to.
        :param loader: a troj batch iterator.
        :param num_batches: How many batches to evaluate. If none, all are evaluated.
        '''
        for idx in range(len(self.callbacks)):
            for batch_idx, (data, target, index) in enumerate(loader):
                if verbose:
                    verb_str = 'Batch {}/{} for callback {}/{}'
                    print(verb_str.format(batch_idx, len(loader), idx+1, len(self.callbacks)))
                if self.model != None:
                    log_index, log_dict, = self.callbacks[idx](self.model, data, target, index, **self.callback_kwargs[idx])
                else:
                    log_index, log_dict, = self.callbacks[idx](data, target, index, **self.callback_kwargs[idx])
                df = trojai.data.log_to_dataframe(df, log_index, log_dict)
                if num_batches is not None and batch_idx > num_batches:
                    break
        return df




import numpy as np
import operator

def get_pred_data(preds):
    """

    :param preds: predictions from object detection model.
    :return: the boxes, labels, and scores as tensors.
    """
    boxes = preds["boxes"].detach().cpu().numpy()
    labels = preds["labels"].detach().cpu().numpy()
    scores = preds["scores"].detach().cpu().numpy()
    return boxes, labels, scores


def compute_Lp_distance(x1, x2, p=np.inf):
    """
    compute Lp distance of a collection of images against another collection.

    :param x1: image collection 1
    :param x2: image collection 2
    :param p: p norm
    :return: tensor of distances
    """
    x1 = np.reshape(x1, (x1.shape[0], -1))
    x2 = np.reshape(x2, (x2.shape[0], -1))
    difference_vect = x1 - x2
    lp_distance = np.linalg.norm(difference_vect, p, axis=1)
    return lp_distance

#TODO document
def SwapKeys(old_dict, key_dict):
    '''
    Swaps the keys in old_dict to their values in key_dict.
    :param old_dict: Dictionary to be assigned new keys.
    :param key_dict: dictionary containing the same keys as old_dict, where the values are the new key values for old_dict.
    :return: A dictionary with the new key names.
    '''
    pretty_dict = {key_dict[k]: v for k, v in old_dict.items()}
    return pretty_dict

def normalize_dicts(numer_dict, denom_dict):
    '''
    Performs elementwise division between two dictionaries.
    :param numer_dict: Dictionary to divide.
    :param denom_dict: Dictionary to divide by.
    :return: Dictionary with divided values.
    '''
    new_dict = {}
    for key in list(numer_dict.keys()):
        new_dict[key] = numer_dict[key]/denom_dict[key]
    return new_dict

def sort_dict(inp):
    '''
    Sorts dictionary alphabetically.
    :param inp: Input dictionary.
    :return: Sorted dictionary.
    '''
    sorted_dict = {}
    for key in sorted(inp):
        sorted_dict[key] = inp[key]
    return sorted_dict

def add_dicts(dict_1, dict_2):
    '''
    Performs elementwise addition between two dictionaries.
    :param dict_1: First dictionary.
    :param dict_2: Second dictionary.
    :return: Elementwise sum of the dictionaries.
    '''
    new_dict = {}
    for key in list(dict_1.keys()):
        new_dict[key] = dict_1[key]+dict_2[key]
    return new_dict

def multiply_dicts(dict_1, dict_2):
    '''
    Performs elementwise multiplication between two dictionaries.
    :param dict_1: First dictionary.
    :param dict_2: Second dictionary.
    :return: Elementwise product of the dictionaries.
    '''
    new_dict = {}
    for key in list(dict_1.keys()):
        new_dict[key] = dict_1[key]*dict_2[key]
    return new_dict

def scalar_mult_dict(scalar, dict_1):
    '''
    Multiplies all values in a dictionary by a given scalar.
    :param scalar: A scalar value.
    :param dict_1: input dictionary
    :return: A dictionary containing the same key-value pairs as dict_1 with the value multiplied by the given scalar.
    '''
    new_dict = {}
    for key in list(dict_1.keys()):
        new_dict[key] = scalar*dict_1[key]
    return new_dict

def diff(first, second):
    '''
    Returns a list of the items in first that are not in second.
    :param first: A list.
    :param second: Another list.
    :return: a list of the items in first that are not in second.
    '''
    second = set(second)
    return [item for item in first if item not in second]

def argmax_dict(vals):
    '''
    Computes an argmax function on a dictionary.
    :param vals: Input dictionary
    :return: The maximum key-value (arg, val) pair as a tuple.
    '''
    arg = max(vals.items(), key=operator.itemgetter(1))[0]
    arg_val = vals[arg]
    return (arg, arg_val)

def argmin_dict(vals):
    '''
    Computes an argmin function on a dictionary.
    :param vals: Input dictionary
    :return: The minimum key-value (arg, val) pair as a tuple.
    '''
    arg = min(vals.items(), key=operator.itemgetter(1))[0]
    arg_val = vals[arg]
    return (arg, arg_val)

def sub_dicts(dict_1, dict_2):
    '''
    Performs elementwise subtraction between two dictionaries.
    :param dict_1: First dictionary.
    :param dict_2: Second dictionary.
    :return: Elementwise difference of the dictionaries.
    '''
    new_dict = {}
    for key in list(dict_1.keys()):
        new_dict[key] = dict_1[key]-dict_2[key]
    return new_dict

def ColumnMeanStd(df, column_name):
    '''
    Computes the mean and deviation of a dataframe column which has scalar values.

    :param df: Dataframe
    :param column_name: The name of the column to compute the values for.
    :return: The mean and standard deviation.
    '''
    vals = df[column_name].tolist()
    mean = round(np.mean(vals), 3)
    dev = round(np.std(vals), 3)
    return mean, dev



import torch
import torch.nn as nn
from collections import defaultdict
import art.estimators.object_detection.python_object_detector

def ODPredsFormatting(predictions,  keys, new_keys):
    '''
    Takes in output of Pytorch object detection, and converts it to the appropriate dictionary format for postprocessing.
    :param predictions: List of output dictionaries from model.
    :param keys: Keys in the old prediction dictionarys.
    :param new_keys: Keys for the new prediction dictionary.
    '''
    preds_dict = defaultdict(list)
    for preds in predictions:
        for key_idx in range(len(keys)):
            key_vals = preds[keys[key_idx]].detach().cpu().numpy()
            key_vals = key_vals.tolist()
            preds_dict[new_keys[key_idx]].append(key_vals)
    return dict(preds_dict)

class PytorchObjectDetector(nn.Module):
    def __init__(self, model, **art_od_kwargs):
        """

        :param model: Pytorch Object Detection model
        """
        super(PytorchObjectDetector, self).__init__()
        self.model = model
        self.art_model = art.estimators.object_detection.python_object_detector.PyTorchObjectDetector(model, **art_od_kwargs)

    def compute_loss(self, samples, labels, grad=False, reduce=True):
        """
        :param: samples: Samples to make predictions on as list of tensors.
        :param: labels: labels as list of dictionaries
        :param: grad: If true, the gradient is computed.
        :param: reduce: If true, the losses are reduced via summation and a single scalar is returned, otherwise a dictionary is returned.
        """
        self.model.train()
        if grad == False:
            with torch.no_grad():
                losses = self.model(samples, labels)
        else:
            losses = self.model(samples, labels)
        if reduce:
            losses = sum(loss for loss in losses.values())
        return losses

    def predict(self, sample):
        self.model.eval()
        preds = self.model(sample)
        return preds


import numpy as np
from queue import Queue
from sklearn.neighbors import NearestNeighbors
import numpy.random

class MemBank:
    #TODO random instantiation
    '''
    Memory bank base class. Allows one to create and update a memory bank with parameter size being the maximum number of entries.
    '''
    def __init__(self, size=4096, input_size=4096):
        '''

        :param size:  Number of entries in the memory bank. Set to 0 for arbitrarily large memory.
        :param input_size: Size of each entry in the bank.
        '''
        self.size = size
        self.bank = Queue(maxsize = size)
        self.input_size = input_size

    def random_init(self):
        '''
        Initializes the memory bank with random vectors.
        :return:
        '''
        for i in range(self.size):
            random = np.random.rand(self.input_size)
            self.bank.put(random)

    def update(self, new_input):
        '''
        Updates the memory bank with new input.
        :param new_input: Input to add to memory bank.
        :return: Adds items to memory bank, removing the oldest entries if the bank is full.
        '''
        for new_item in new_input:
            if self.bank.full() == True:
                self.bank.get()
            self.bank.put(new_item)

    def get_all(self):
        '''

        :return: All elements in the memeory bank as a list.
        '''
        return list(self.bank.queue)


class DimReductionMemBank(MemBank):
    '''
    Memory bank for using dimensionality reduction algorithms
    '''
    def __init__(self, size=0, input_size=128):
        super().__init__(size, input_size)
        self.idx_bank = Queue(maxsize = size)

    def update(self, new_input, new_input_idx):
        '''
        Updates the memory bank with new input.
        :param new_input: Input to add to memory bank.
        :return: Adds items to memory bank, removing the oldest entries if the bank is full.
        '''
        for new_item_ind in range(new_input.shape[0]):
            if self.bank.full() == True:
                self.bank.get()
                self.idx_bank.get()
            self.bank.put(new_input[new_item_ind])
            self.idx_bank.put(new_input_idx[new_item_ind])

    def get_all(self):
        '''

        :return: All elements in the memeory bank as a list.
        '''
        return list(self.bank.queue), list(self.idx_bank.queue)